from typing import (List, Optional, Union)
from typing_extensions import Literal
from pydantic import conint
from enum import Enum

from rastervision.pipeline.config import (ConfigError, register_config, Field,
                                          Config, validator)
from rastervision.core.rv_pipeline.rv_pipeline_config import (RVPipelineConfig,
                                                              PredictOptions)
from rastervision.core.data import SemanticSegmentationLabelStoreConfig
from rastervision.core.evaluation import SemanticSegmentationEvaluatorConfig


class SemanticSegmentationWindowMethod(Enum):
    """Enum for window methods

    Attributes:
        sliding: use a sliding window
        random_sample: randomly sample windows
    """

    sliding = 'sliding'
    random_sample = 'random_sample'


def ss_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version < 1:
        try:
            # removed in version 1
            del cfg_dict['channel_display_groups']
            del cfg_dict['img_format']
            del cfg_dict['label_format']
        except KeyError:
            pass
    return cfg_dict


@register_config('semantic_segmentation_chip_options')
class SemanticSegmentationChipOptions(Config):
    """Chipping options for semantic segmentation."""
    window_method: SemanticSegmentationWindowMethod = Field(
        SemanticSegmentationWindowMethod.sliding,
        description=('Window method to use for chipping.'))
    target_class_ids: Optional[List[int]] = Field(
        None,
        description=
        ('List of class ids considered as targets (ie. those to prioritize when '
         'creating chips) which is only used in conjunction with the '
         'target_count_threshold and negative_survival_probability options. Applies '
         'to the random_sample window method.'))
    negative_survival_prob: float = Field(
        1.0, description='Probability of keeping a negative chip.')
    chips_per_scene: int = Field(
        1000,
        description=
        ('Number of chips to generate per scene. Applies to the random_sample window '
         'method.'))
    target_count_threshold: int = Field(
        1000,
        description=
        ('Minimum number of pixels covering target_classes that a chip must have. '
         'Applies to the random_sample window method.'))
    stride: Optional[int] = Field(
        None,
        description=
        ('Stride of windows across image. Defaults to half the chip size. Applies to '
         'the sliding_window method.'))


@register_config('semantic_segmentation_predict_options')
class SemanticSegmentationPredictOptions(PredictOptions):
    stride: Optional[int] = Field(
        None,
        description=
        'Stride of windows across image. Allows aggregating multiple '
        'predictions for each pixel if less than the chip size. '
        'Defaults to predict_chip_sz.')
    crop_sz: Optional[Union[conint(gt=0), Literal['auto']]] = Field(
        None,
        description=
        'Number of rows/columns of pixels from the edge of prediction '
        'windows to discard. This is useful because predictions near edges '
        'tend to be lower quality and can result in very visible artifacts '
        'near the edges of chips. If "auto", will be set to half the stride '
        'if stride is less than chip_sz. Defaults to None.')

    @validator('crop_sz')
    def validate_crop_sz(cls,
                         v: Optional[Union[conint(gt=0), Literal['auto']]],
                         values: dict) -> dict:
        stride: Optional[int] = values.get('stride')
        crop_sz = v

        if stride is None and crop_sz is not None:
            raise ConfigError('Cannot use crop_sz if stride is None.')

        return crop_sz


@register_config('semantic_segmentation', upgrader=ss_config_upgrader)
class SemanticSegmentationConfig(RVPipelineConfig):
    """Configure a :class:`.SemanticSegmentation` pipeline."""

    chip_options: SemanticSegmentationChipOptions = \
        SemanticSegmentationChipOptions()
    predict_options: SemanticSegmentationPredictOptions = \
        SemanticSegmentationPredictOptions()

    def build(self, tmp_dir):
        from rastervision.core.rv_pipeline.semantic_segmentation import (
            SemanticSegmentation)
        return SemanticSegmentation(self, tmp_dir)

    def update(self):
        self.dataset.class_config.ensure_null_class()
        super().update()

    def validate_config(self):
        super().validate_config()

    def get_default_label_store(self, scene):
        return SemanticSegmentationLabelStoreConfig()

    def get_default_evaluator(self):
        return SemanticSegmentationEvaluatorConfig()



from enum import Enum
from typing import Optional

from rastervision.pipeline.config import register_config, Config, Field
from rastervision.core.rv_pipeline import RVPipelineConfig, PredictOptions
from rastervision.core.data.label_store import ObjectDetectionGeoJSONStoreConfig
from rastervision.core.evaluation import ObjectDetectionEvaluatorConfig


class ObjectDetectionWindowMethod(Enum):
    """Enum for window methods

    Attributes:
        chip: the default method
    """
    chip = 'chip'
    label = 'label'
    image = 'image'
    sliding = 'sliding'


@register_config('object_detection_chip_options')
class ObjectDetectionChipOptions(Config):
    neg_ratio: float = Field(
        1.0,
        description=
        ('The ratio of negative chips (those containing no bounding '
         'boxes) to positive chips. This can be useful if the statistics '
         'of the background is different in positive chips. For example, '
         'in car detection, the positive chips will always contain roads, '
         'but no examples of rooftops since cars tend to not be near rooftops.'
         ))
    ioa_thresh: float = Field(
        0.8,
        description=
        ('When a box is partially outside of a training chip, it is not clear if (a '
         'clipped version) of the box should be included in the chip. If the IOA '
         '(intersection over area) of the box with the chip is greater than ioa_thresh, '
         'it is included in the chip.'))
    window_method: ObjectDetectionWindowMethod = ObjectDetectionWindowMethod.chip
    label_buffer: Optional[int] = None


@register_config('object_detection_predict_options')
class ObjectDetectionPredictOptions(PredictOptions):
    merge_thresh: float = Field(
        0.5,
        description=
        ('If predicted boxes have an IOA (intersection over area) greater than '
         'merge_thresh, then they are merged into a single box during postprocessing. '
         'This is needed since the sliding window approach results in some false '
         'duplicates.'))
    score_thresh: float = Field(
        0.5,
        description=
        ('Predicted boxes are only output if their score is above score_thresh.'
         ))


@register_config('object_detection')
class ObjectDetectionConfig(RVPipelineConfig):
    """Configure an :class:`.ObjectDetection` pipeline."""

    chip_options: ObjectDetectionChipOptions = ObjectDetectionChipOptions()
    predict_options: ObjectDetectionPredictOptions = ObjectDetectionPredictOptions(
    )

    def build(self, tmp_dir):
        from rastervision.core.rv_pipeline.object_detection import ObjectDetection
        return ObjectDetection(self, tmp_dir)

    def get_default_label_store(self, scene):
        return ObjectDetectionGeoJSONStoreConfig()

    def get_default_evaluator(self):
        return ObjectDetectionEvaluatorConfig()



from typing import TYPE_CHECKING, List, Sequence
import logging

import numpy as np

from rastervision.core.box import Box
from rastervision.core.rv_pipeline.rv_pipeline import RVPipeline
from rastervision.core.rv_pipeline.utils import nodata_below_threshold
from rastervision.core.rv_pipeline.semantic_segmentation_config import (
    SemanticSegmentationWindowMethod)

if TYPE_CHECKING:
    from rastervision.core.backend.backend import Backend
    from rastervision.core.data import (
        ClassConfig,
        Labels,
        Scene,
        SemanticSegmentationLabelSource,
    )
    from rastervision.core.rv_pipeline.semantic_segmentation_config import (
        SemanticSegmentationConfig, SemanticSegmentationChipOptions)

log = logging.getLogger(__name__)


def get_train_windows(scene: 'Scene',
                      class_config: 'ClassConfig',
                      chip_size: int,
                      chip_options: 'SemanticSegmentationChipOptions',
                      chip_nodata_threshold: float = 1.) -> List[Box]:
    """Get training windows covering a scene.

    Args:
        scene: The scene over-which windows are to be generated.

    Returns:
        A list of windows, list(Box)
    """
    co = chip_options
    raster_source = scene.raster_source
    extent = raster_source.extent
    label_source: 'SemanticSegmentationLabelSource' = scene.label_source

    def filter_windows(windows: Sequence[Box]) -> List[Box]:
        """Filter out chips that
        (1) are outside the AOI
        (2) only consist of null labels
        (3) have NODATA proportion >= chip_nodata_threshold
        """
        total_windows = len(windows)
        if scene.aoi_polygons:
            windows = Box.filter_by_aoi(windows, scene.aoi_polygons)
            log.info(f'AOI filtering: {len(windows)}/{total_windows} '
                     'chips accepted')

        filt_windows = []
        for w in windows:
            chip = raster_source.get_chip(w)
            nodata_below_thresh = nodata_below_threshold(
                chip, chip_nodata_threshold, nodata_val=0)

            label_arr = label_source.get_labels(w).get_label_arr(w)
            null_labels = label_arr == class_config.null_class_id

            if not np.all(null_labels) and nodata_below_thresh:
                filt_windows.append(w)
        log.info('Label and NODATA filtering: '
                 f'{len(filt_windows)}/{len(windows)} chips accepted')

        windows = filt_windows
        return windows

    def should_use_window(window: Box) -> bool:
        if co.negative_survival_prob >= 1.0:
            return True
        else:
            is_positive = False
            if co.target_class_ids is not None:
                is_positive = label_source.enough_target_pixels(
                    window, co.target_count_threshold, co.target_class_ids)
            if is_positive:
                return True
            keep_negative = np.random.sample() < co.negative_survival_prob
            return keep_negative

    if co.window_method == SemanticSegmentationWindowMethod.sliding:
        stride = co.stride or int(round(chip_size / 2))
        unfiltered_windows = extent.get_windows(chip_size, stride)
        windows = list(filter_windows(unfiltered_windows))
        if len(windows) > 0:
            a_window = windows[0]
            windows = list(filter(should_use_window, windows))
            if len(windows) == 0:
                windows = [a_window]
        elif len(windows) == 0:
            return [unfiltered_windows[0]]
    elif co.window_method == SemanticSegmentationWindowMethod.random_sample:
        windows = []
        attempts = 0

        while attempts < co.chips_per_scene:
            window = extent.make_random_square(chip_size)
            if not filter_windows([window]):
                continue

            attempts += 1
            if co.negative_survival_prob >= 1.0:
                windows.append(window)
            elif attempts == co.chips_per_scene and len(windows) == 0:
                # Ensure there is at least one window per scene.
                windows.append(window)
            elif should_use_window(window):
                windows.append(window)

    return windows


class SemanticSegmentation(RVPipeline):
    def chip(self, *args, **kwargs):
        log.info(f'Chip options: {self.config.chip_options}')
        return super().chip(*args, **kwargs)

    def get_train_windows(self, scene):
        return get_train_windows(
            scene,
            self.config.dataset.class_config,
            self.config.train_chip_sz,
            self.config.chip_options,
            chip_nodata_threshold=self.config.chip_nodata_threshold)

    def get_train_labels(self, window, scene):
        return scene.label_source.get_labels(window=window)

    def post_process_batch(self, windows, chips, labels):
        # Fill in null class for any NODATA pixels.
        null_class_id = self.config.dataset.class_config.null_class_id
        for window, chip in zip(windows, chips):
            nodata_mask = np.sum(chip, axis=2) == 0
            labels.mask_fill(window, nodata_mask, fill_value=null_class_id)

        return labels

    def predict_scene(self, scene: 'Scene', backend: 'Backend') -> 'Labels':
        cfg: 'SemanticSegmentationConfig' = self.config
        chip_sz = cfg.predict_chip_sz
        stride = cfg.predict_options.stride
        crop_sz = cfg.predict_options.crop_sz

        if stride is None:
            stride = chip_sz

        if crop_sz == 'auto':
            overlap_sz = chip_sz - stride
            if overlap_sz % 2 == 1:
                log.warn(
                    'Using crop_sz="auto" but overlap size (chip_sz minus '
                    'stride) is odd. This means that one pixel row/col will '
                    'still overlap after cropping.')
            crop_sz = overlap_sz // 2

        return backend.predict_scene(
            scene, chip_sz=chip_sz, stride=stride, crop_sz=crop_sz)



from typing import TYPE_CHECKING

import numpy as np
from sklearn.metrics import confusion_matrix

from rastervision.core.evaluation import (ClassificationEvaluation,
                                          ClassEvaluationItem)
if TYPE_CHECKING:
    from rastervision.core.data import (ChipClassificationLabels, ClassConfig)


class ChipClassificationEvaluation(ClassificationEvaluation):
    def __init__(self, class_config: 'ClassConfig'):
        super().__init__()
        self.class_config = class_config

    def compute(self, gt_labels: 'ChipClassificationLabels',
                pred_labels: 'ChipClassificationLabels') -> None:
        self.reset()
        self.class_to_eval_item = {}

        gt_class_ids = []
        pred_class_ids = []
        for gt_cell in gt_labels.get_cells():
            gt_class_id = gt_labels.get_cell_class_id(gt_cell)
            pred_class_id = pred_labels.get_cell_class_id(gt_cell)

            if gt_class_id is not None and pred_class_id is not None:
                gt_class_ids.append(gt_class_id)
                pred_class_ids.append(pred_class_id)

        labels = np.arange(len(self.class_config))
        self.conf_mat = confusion_matrix(
            gt_class_ids, pred_class_ids, labels=labels)

        for class_id, class_name in enumerate(self.class_config.names):
            eval_item = ClassEvaluationItem.from_multiclass_conf_mat(
                conf_mat=self.conf_mat,
                class_id=class_id,
                class_name=class_name)
            self.class_to_eval_item[class_id] = eval_item

        self.compute_avg()



from typing import Optional

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.raster_transformer.raster_transformer_config import (  # noqa
    RasterTransformerConfig)
from rastervision.core.data.raster_transformer.nan_transformer import (  # noqa
    NanTransformer)


@register_config('nan_transformer')
class NanTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.NanTransformer`."""

    to_value: Optional[float] = Field(
        0.0, description=('Turn all NaN values into this value.'))

    def update(self, pipeline=None, scene=None):
        if pipeline is not None and self.to_value is None:
            self.to_value = pipeline.to_value

    def build(self):
        return NanTransformer(to_value=self.to_value)



from typing import Optional

from rastervision.core.data.vector_source import (VectorSourceConfig)
from rastervision.core.data.label_source import (LabelSourceConfig,
                                                 ChipClassificationLabelSource)
from rastervision.pipeline.config import (ConfigError, register_config, Field,
                                          validator, root_validator)
from rastervision.core.data.vector_transformer import (
    ClassInferenceTransformerConfig, BufferTransformerConfig)


def cc_label_source_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 4:
        # made non-optional in version 5
        cfg_dict['ioa_thresh'] = cfg_dict.get('ioa_thresh', 0.5)
    return cfg_dict


@register_config(
    'chip_classification_label_source',
    upgrader=cc_label_source_config_upgrader)
class ChipClassificationLabelSourceConfig(LabelSourceConfig):
    """Configure a :class:`.ChipClassificationLabelSource`.

    This can be provided explicitly as a grid of cells, or a grid of cells can
    be inferred from arbitrary polygons.
    """
    vector_source: Optional[VectorSourceConfig] = None
    ioa_thresh: float = Field(
        0.5,
        description=
        ('Minimum IOA of a polygon and cell for that polygon to be a candidate for '
         'setting the class_id.'))
    use_intersection_over_cell: bool = Field(
        False,
        description=
        ('If True, then use the area of the cell as the denominator in the IOA. '
         'Otherwise, use the area of the polygon.'))
    pick_min_class_id: bool = Field(
        False,
        description=
        ('If True, the class_id for a cell is the minimum class_id of the boxes in that '
         'cell. Otherwise, pick the class_id of the box covering the greatest area.'
         ))
    background_class_id: Optional[int] = Field(
        None,
        description=
        ('If not None, class_id to use as the background class; ie. the one that is used '
         'when a window contains no boxes. Cannot be None if infer_cells=True.'
         ))
    infer_cells: bool = Field(
        False,
        description='If True, infers a grid of cells based on the cell_sz.')
    cell_sz: Optional[int] = Field(
        None,
        description=
        ('Size of a cell to use in pixels. If None, and this Config is part '
         'of an RVPipeline, this field will be set from RVPipeline.train_chip_sz.'
         ))
    lazy: bool = Field(
        False,
        description='If True, labels will not be populated automatically '
        'during initialization of the label source.')

    @validator('vector_source')
    def ensure_required_transformers(
            cls, v: VectorSourceConfig) -> VectorSourceConfig:
        """Add class-inference and buffer transformers if absent."""
        tfs = v.transformers

        # add class inference transformer
        has_inf_tf = any(
            isinstance(tf, ClassInferenceTransformerConfig) for tf in tfs)
        if not has_inf_tf:
            tfs += [ClassInferenceTransformerConfig(default_class_id=None)]

        # add buffer transformers
        has_buf_tf = any(isinstance(tf, BufferTransformerConfig) for tf in tfs)
        if not has_buf_tf:
            tfs += [
                BufferTransformerConfig(geom_type='Point', default_buf=1),
                BufferTransformerConfig(geom_type='LineString', default_buf=1)
            ]

        return v

    @root_validator(skip_on_failure=True)
    def ensure_bg_class_id_if_inferring(cls, values: dict) -> dict:
        infer_cells = values.get('infer_cells')
        has_bg_class_id = values.get('background_class_id') is not None
        if infer_cells and not has_bg_class_id:
            raise ConfigError(
                'background_class_id is required if infer_cells=True.')
        return values

    def build(self, class_config, crs_transformer, extent=None, tmp_dir=None):
        if self.vector_source is None:
            raise ValueError('Cannot build with a None vector_source.')
        if self.infer_cells and self.cell_sz is None and not self.lazy:
            raise ValueError('Cannot build with infer_cells=True, '
                             'cell_sz=None and lazy=True.')
        vector_source = self.vector_source.build(class_config, crs_transformer)
        return ChipClassificationLabelSource(
            self, vector_source, extent=extent, lazy=self.lazy)

    def update(self, pipeline=None, scene=None):
        super().update(pipeline, scene)
        if self.cell_sz is None and pipeline is not None:
            self.cell_sz = pipeline.train_chip_sz
        if self.vector_source is not None:
            self.vector_source.update(pipeline, scene)



from typing import TYPE_CHECKING, List
import logging

from rasterio.features import rasterize
import numpy as np
import geopandas as gpd

from rastervision.core.data.raster_source import RasterSource

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rastervision.core.box import Box
    from rastervision.core.data import VectorSource, RasterTransformer


def geoms_to_raster(df: gpd.GeoDataFrame, window: 'Box',
                    background_class_id: int, all_touched: bool) -> np.ndarray:
    """Rasterize geometries that intersect with the window.

    Args:
        df (gpd.GeoDataFrame): All label geometries in the scene.
        window (Box): The part of the scene to rasterize.
        background_class_id (int): Class ID to use for pixels that don't
            fall under any label geometry.
        all_touched (bool): If True, all pixels touched by geometries will be
            burned in. If false, only pixels whose center is within the
            polygon or that are selected by Bresenham's line algorithm will be
            burned in. (See :func:`.rasterize` for more details).
            Defaults to False.

    Returns:
        np.ndarray: A raster.
    """
    if len(df) == 0:
        return np.full(window.size, background_class_id, dtype=np.uint8)

    window_geom = window.to_shapely()

    # subset to shapes that intersect window
    df_int = df[df.intersects(window_geom)]
    # transform to window frame of reference
    shapes = df_int.translate(xoff=-window.xmin, yoff=-window.ymin)
    # class IDs of each shape
    class_ids = df_int['class_id']

    if len(shapes) > 0:
        raster = rasterize(
            shapes=list(zip(shapes, class_ids)),
            out_shape=window.size,
            fill=background_class_id,
            dtype=np.uint8,
            all_touched=all_touched)
    else:
        raster = np.full(window.size, background_class_id, dtype=np.uint8)

    return raster


class RasterizedSource(RasterSource):
    """A :class:`.RasterSource` based on the rasterization of a VectorSource."""

    def __init__(self,
                 vector_source: 'VectorSource',
                 background_class_id: int,
                 extent: 'Box',
                 all_touched: bool = False,
                 raster_transformers: List['RasterTransformer'] = []):
        """Constructor.

        Args:
            vector_source (VectorSource): The VectorSource to rasterize.
            background_class_id (int): The class_id to use for any background
                pixels, ie. pixels not covered by a polygon.
            extent (Box): Extent of corresponding imagery RasterSource.
            all_touched (bool, optional): If True, all pixels touched by
                geometries will be burned in. If false, only pixels whose
                center is within the polygon or that are selected by
                Bresenham's line algorithm will be burned in.
                (See :func:`~rasterio.features.rasterize` for more details).
                Defaults to False.
        """
        self.vector_source = vector_source
        self.background_class_id = background_class_id
        self.all_touched = all_touched

        self.df = self.vector_source.get_dataframe()
        self.validate_labels(self.df)

        super().__init__(
            channel_order=[0],
            num_channels_raw=1,
            raster_transformers=raster_transformers,
            extent=extent)

    @property
    def dtype(self) -> np.dtype:
        return np.uint8

    @property
    def crs_transformer(self):
        return self.vector_source.crs_transformer

    def _get_chip(self, window: 'Box') -> np.ndarray:
        """Return the chip located in the window.

        Polygons falling within the window are rasterized using their
        ``class_id`` property and the background is filled with
        ``background_class_id``.

        Args:
            window (Box): Window to read.

        Returns:
            np.ndarray: [height, width, channels] numpy array
        """
        log.debug(f'Rasterizing window: {window}')
        chip = geoms_to_raster(
            self.df,
            window,
            background_class_id=self.background_class_id,
            all_touched=self.all_touched)
        # Add third singleton dim since rasters must have >=1 channel.
        return np.expand_dims(chip, 2)

    def validate_labels(self, df: gpd.GeoDataFrame) -> None:
        """Validate label geometries.

        Args:
            df (gpd.GeoDataFrame): Label geometries.

        Raises:
            ValueError: If ``Point`` or ``LineString`` geometries found.
            ValueError: If geometries are missing class IDs.
        """
        geom_types = set(df.geom_type)
        if 'Point' in geom_types or 'LineString' in geom_types:
            raise ValueError('LineStrings and Points are not supported '
                             'in RasterizedSource. Use BufferTransformer '
                             'to buffer them into Polygons. '
                             f'Geom types found in data: {geom_types}')

        if len(df) > 0 and 'class_id' not in df.columns:
            raise ValueError('All label polygons must have a class_id.')



from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
from PIL import ImageColor


def color_to_triple(
        color: Optional[Union[str, Sequence]] = None) -> Tuple[int, int, int]:
    """Given a PIL ImageColor string, return a triple of integers
    representing the red, green, and blue values.

    If color is None, return a random color.

    Args:
         color: A PIL ImageColor string

    Returns:
         An triple of integers

    """
    if color is None:
        r, g, b = np.random.randint(0, 256, size=3).tolist()
        return r, g, b
    elif isinstance(color, str):
        return ImageColor.getrgb(color)
    elif isinstance(color, (tuple, list)):
        return color
    else:
        raise TypeError(f'Unsupported type: {type(color)}')


def color_to_integer(color: str) -> int:
    """Given a PIL ImageColor string, return a packed integer.

    Args:
         color: A PIL ImageColor string

    Returns:
         An integer containing the packed RGB values.

    """
    triple = color_to_triple(color)
    r = triple[0] * (1 << 16)
    g = triple[1] * (1 << 8)
    b = triple[2] * (1 << 0)
    integer = r + g + b
    return integer


def normalize_color(
        color: Union[str, tuple, list]) -> Tuple[float, float, float]:
    """Convert color representation to a float 3-tuple with values in [0-1]."""
    if isinstance(color, str):
        color = color_to_triple(color)

    if isinstance(color, (tuple, list)):
        if all(isinstance(c, int) for c in color):
            return tuple(c / 255. for c in color)
        elif all(isinstance(c, float) for c in color):
            return tuple(color)
        else:
            raise ValueError('RGB values must be either all ints (0-255) '
                             'or all floats (0.0-1.0)')

    raise TypeError('Expected color to be a string or tuple or list, '
                    f'but found {type(color)}.')


def rgb_to_int_array(rgb_array: np.ndarray) -> np.ndarray:
    r = np.array(rgb_array[:, :, 0], dtype=np.uint32) * (1 << 16)
    g = np.array(rgb_array[:, :, 1], dtype=np.uint32) * (1 << 8)
    b = np.array(rgb_array[:, :, 2], dtype=np.uint32) * (1 << 0)
    return r + g + b


def all_equal(it: list):
    ''' Returns true if all elements are equal to each other '''
    return it.count(it[0]) == len(it)


def listify_uris(uris: Union[str, List[str]]) -> List[str]:
    """Convert to URI to list if needed."""
    if isinstance(uris, (list, tuple)):
        pass
    elif isinstance(uris, str):
        uris = [uris]
    else:
        raise TypeError(f'Expected str or List[str], but got {type(uris)}.')
    return uris



from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.class_config import ClassConfig
from rastervision.core.data.raster_transformer import (RasterTransformerConfig,
                                                       RGBClassTransformer)


@register_config('rgb_class_transformer')
class RGBClassTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.RGBClassTransformer`."""

    class_config: ClassConfig = Field(
        ...,
        description=('The class config defining the mapping between '
                     'classes and colors.'))

    def build(self) -> RGBClassTransformer:
        return RGBClassTransformer(class_config=self.class_config)



from typing import TYPE_CHECKING, Optional

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.vector_transformer import (VectorTransformerConfig,
                                                       ShiftTransformer)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


@register_config('shift_transformer')
class ShiftTransformerConfig(VectorTransformerConfig):
    """Configure a :class:`.ShiftTransformer`."""

    x_shift: float = Field(
        0.0,
        descriptions='Distance in meters to shift along the x-axis. '
        'Postive values shift eastward.')
    y_shift: float = Field(
        0.0,
        descriptions='Distance in meters to shift along the y-axis. '
        'Postive values shift northward.')
    round_pixels: bool = Field(
        True,
        descriptions='Whether to round shifted pixel values to integers.')

    def build(self, class_config: Optional['ClassConfig'] = None
              ) -> ShiftTransformer:
        return ShiftTransformer(
            x_shift=self.x_shift,
            y_shift=self.y_shift,
            round_pixels=self.round_pixels)



from typing import Optional
from os.path import join

from rastervision.core.data.label_store import (LabelStoreConfig,
                                                ObjectDetectionGeoJSONStore)
from rastervision.pipeline.config import register_config, Field


@register_config('object_detection_geojson_store')
class ObjectDetectionGeoJSONStoreConfig(LabelStoreConfig):
    """Configure an :class:`.ObjectDetectionGeoJSONStore`."""

    uri: Optional[str] = Field(
        None,
        description=
        ('URI of GeoJSON file with predictions. If None, and this Config is part of '
         'a SceneConfig inside an RVPipelineConfig, it will be auto-generated.'
         ))

    def build(self, class_config, crs_transformer, extent=None, tmp_dir=None):
        return ObjectDetectionGeoJSONStore(self.uri, class_config,
                                           crs_transformer)

    def update(self, pipeline=None, scene=None):
        if pipeline is not None and scene is not None:
            if self.uri is None:
                self.uri = join(pipeline.predict_uri,
                                '{}.json'.format(scene.id))



from typing import TYPE_CHECKING, Dict, Optional

from rastervision.core.data.utils.geojson import buffer_geoms
from rastervision.core.data.vector_transformer import VectorTransformer

if TYPE_CHECKING:
    from rastervision.core.data import CRSTransformer


class BufferTransformer(VectorTransformer):
    """Buffers geometries."""

    def __init__(self,
                 geom_type: str,
                 class_bufs: Optional[Dict[int, Optional[float]]] = None,
                 default_buf: Optional[float] = None):
        """Constructor.

        Args:
            geom_type (str): The geometry type to apply this transform to.
                E.g. "LineString", "Point", "Polygon".
            class_bufs (Dict[int, Optional[float]], optional): Mapping from
                class IDs to buffer amounts (in pixels). If a class ID is not
                found in the mapping, the value specified by the default_buf
                field will be used. If the buffer value for a class is None,
                then no buffering will be applied to the geoms of that
                class.
                Defaults to {}.
            default_buf (Optional[float], optional): Default buffer to apply to
                classes not in class_bufs. If None, no buffering will be
                applied to the geoms of those missing classes. Defaults to
                None.
        """
        self.geom_type = geom_type
        self.class_bufs = class_bufs if class_bufs is not None else {}
        self.default_buf = default_buf

    def transform(self,
                  geojson: dict,
                  crs_transformer: Optional['CRSTransformer'] = None) -> dict:
        return buffer_geoms(
            geojson,
            self.geom_type,
            class_bufs=self.class_bufs,
            default_buf=self.default_buf)



# flake8: noqa

from rastervision.core.data.label_store.label_store import *
from rastervision.core.data.label_store.label_store_config import *
from rastervision.core.data.label_store.chip_classification_geojson_store import *
from rastervision.core.data.label_store.chip_classification_geojson_store_config import *
from rastervision.core.data.label_store.semantic_segmentation_label_store import *
from rastervision.core.data.label_store.semantic_segmentation_label_store_config import *
from rastervision.core.data.label_store.object_detection_geojson_store import *
from rastervision.core.data.label_store.object_detection_geojson_store_config import *
from rastervision.core.data.label_store.utils import *

__all__ = [
    LabelStore.__name__,
    LabelStoreConfig.__name__,
    SemanticSegmentationLabelStore.__name__,
    SemanticSegmentationLabelStoreConfig.__name__,
    VectorOutputConfig.__name__,
    BuildingVectorOutputConfig.__name__,
    PolygonVectorOutputConfig.__name__,
    ChipClassificationGeoJSONStore.__name__,
    ChipClassificationGeoJSONStoreConfig.__name__,
    ObjectDetectionGeoJSONStore.__name__,
    ObjectDetectionGeoJSONStoreConfig.__name__,
    boxes_to_geojson.__name__,
]



from typing import TYPE_CHECKING, List, Optional, Sequence

import numpy as np

from rastervision.core.data.raster_transformer import RasterTransformer

if TYPE_CHECKING:
    from rastervision.core.data import RasterSource


class StatsTransformer(RasterTransformer):
    """Transforms non-uint8 to uint8 values using channel statistics.

    This works as follows:

    - Convert pixel values to z-scores using channel means and standard
      deviations.
    - Clip z-scores to the specified number of standard deviations (default 3)
      on each side.
    - Scale values to 0-255 and cast to uint8.

    This transformation is not applied to NODATA pixels (assumed to be pixels
    with all values equal to zero).
    """

    def __init__(self,
                 means: Sequence[float],
                 stds: Sequence[float],
                 max_stds: float = 3.):
        """Construct a new StatsTransformer.

        Args:
            means (np.ndarray): Channel means.
            means (np.ndarray): Channel standard deviations.
            max_stds (float): Number of standard deviations to clip the
                distribution to on both sides. Defaults to 3.
        """
        # shape = (1, 1, num_channels)
        self.means = np.array(means, dtype=float)[np.newaxis, np.newaxis, :]
        self.stds = np.array(stds, dtype=float)[np.newaxis, np.newaxis, :]
        self.max_stds = max_stds

    def transform(self,
                  chip: np.ndarray,
                  channel_order: Optional[Sequence[int]] = None) -> np.ndarray:
        """Transform a chip.

        Transforms non-uint8 to uint8 values using raster_stats.

        Args:
            chip: ndarray of shape [height, width, channels] This is assumed to already
                have the channel_order applied to it if channel_order is set. In other
                words, channels should be equal to len(channel_order).
            channel_order: list of indices of channels that were extracted from the
                raw imagery.

        Returns:
            [height, width, channels] uint8 numpy array

        """
        if chip.dtype == np.uint8:
            return chip

        means = self.means
        stds = self.stds
        max_stds = self.max_stds
        if channel_order is not None:
            means = means[..., channel_order]
            stds = stds[..., channel_order]

        # Don't transform NODATA zero values.
        nodata_mask = chip == 0

        # Subtract mean and divide by std to get zscores.
        chip = chip.astype(float)
        chip -= means
        chip /= stds

        # Make zscores that fall between -max_stds and max_stds span 0 to 255.
        # range: (-max_stds, max_stds)
        chip = np.clip(chip, -max_stds, max_stds, out=chip)
        # range: [0, 2 * max_stds]
        chip += max_stds
        # range: [0, 1]
        chip /= (2 * max_stds)
        # range: [0, 255]
        chip *= 255
        chip = chip.astype(np.uint8)

        chip[nodata_mask] = 0

        return chip

    @classmethod
    def from_raster_sources(cls,
                            raster_sources: List['RasterSource'],
                            sample_prob: float = 0.1,
                            max_stds: float = 3.) -> 'StatsTransformer':
        """Create a StatsTransformer with stats from the given raster sources.

        Args:
            raster_sources (List[&#39;RasterSource&#39;]): List of raster
                sources to compute stats from.
            sample_prob (float, optional): Fraction of each raster to sample
                for computing stats. For details see docs for
                RasterStats.compute(). Defaults to 0.1.
            max_stds (float, optional): Number of standard deviations to clip
                the distribution to on both sides. Defaults to 3.

        Returns:
            StatsTransformer: A StatsTransformer.
        """
        from rastervision.core.raster_stats import RasterStats
        stats = RasterStats()
        stats.compute(raster_sources=raster_sources, sample_prob=sample_prob)
        stats_transformer = StatsTransformer(
            means=stats.means, stds=stats.stds, max_stds=max_stds)
        return stats_transformer



import io
from pydantic import confloat

from PIL import Image
import numpy as np
import imageio
import logging

Proportion = confloat(ge=0, le=1)

log = logging.getLogger(__name__)


def save_img(im_array, output_path):
    imageio.imwrite(output_path, im_array)


def numpy_to_png(array: np.ndarray) -> str:
    """Get a PNG string from a Numpy array.

    Args:
         array: A Numpy array of shape (w, h, 3) or (w, h), where the
               former is meant to become a three-channel image and the
               latter a one-channel image.  The dtype of the array
               should be uint8.

    Returns:
         str

    """
    im = Image.fromarray(array)
    output = io.BytesIO()
    im.save(output, 'png')
    return output.getvalue()


def png_to_numpy(png: str, dtype=np.uint8) -> np.ndarray:
    """Get a Numpy array from a PNG string.

    Args:
         png: A str containing a PNG-formatted image.

    Returns:
         numpy.ndarray

    """
    incoming = io.BytesIO(png)
    im = Image.open(incoming)
    return np.array(im)



from typing import Optional, Sequence, List, Tuple
from pydantic import conint

import numpy as np

from rastervision.core.box import Box
from rastervision.core.data.raster_source import RasterSource
from rastervision.core.data.crs_transformer import CRSTransformer
from rastervision.core.data.raster_source.rasterio_source import RasterioSource
from rastervision.core.data.utils import all_equal


class MultiRasterSource(RasterSource):
    """Merge multiple ``RasterSources`` by concatenating along channel dim."""

    def __init__(self,
                 raster_sources: Sequence[RasterSource],
                 primary_source_idx: conint(ge=0) = 0,
                 force_same_dtype: bool = False,
                 channel_order: Optional[Sequence[conint(ge=0)]] = None,
                 raster_transformers: Sequence = [],
                 extent: Optional[Box] = None):
        """Constructor.

        Args:
            raster_sources (Sequence[RasterSource]): Sequence of RasterSources.
            primary_source_idx (0 <= int < len(raster_sources)): Index of the
                raster source whose CRS, dtype, and other attributes will
                override those of the other raster sources.
            force_same_dtype (bool): If true, force all sub-chips to have the
                same dtype as the primary_source_idx-th sub-chip. No careful
                conversion is done, just a quick cast. Use with caution.
            channel_order (Sequence[conint(ge=0)], optional): Channel ordering
                that will be used by .get_chip(). Defaults to None.
            raster_transformers (Sequence, optional): Sequence of transformers.
                Defaults to [].
        """
        num_channels_raw = sum(rs.num_channels_raw for rs in raster_sources)
        if not channel_order:
            num_channels = sum(rs.num_channels for rs in raster_sources)
            channel_order = list(range(num_channels))

        # validate primary_source_idx
        if not (0 <= primary_source_idx < len(raster_sources)):
            raise IndexError('primary_source_idx must be in range '
                             '[0, len(raster_sources)].')

        if extent is None:
            extent = raster_sources[primary_source_idx].extent

        super().__init__(
            channel_order,
            num_channels_raw,
            raster_transformers=raster_transformers,
            extent=extent)

        self.force_same_dtype = force_same_dtype
        self.raster_sources = raster_sources
        self.primary_source_idx = primary_source_idx

        self.extents = [rs.extent for rs in self.raster_sources]
        self.all_extents_equal = all_equal(self.extents)

        self.validate_raster_sources()

    def validate_raster_sources(self) -> None:
        """Validate sub-``RasterSources``.

        Checks if:

        - dtypes are same or ``force_same_dtype`` is True.
        - each sub-``RasterSource`` is a :class:`.RasterioSource` if extents
          not identical.

        """
        dtypes = [rs.dtype for rs in self.raster_sources]
        if not self.force_same_dtype and not all_equal(dtypes):
            raise ValueError(
                'dtypes of all sub raster sources must be the same. '
                f'Got: {dtypes} '
                '(Use force_same_dtype to cast all to the dtype of the '
                'primary source)')
        if not self.all_extents_equal:
            all_rasterio_sources = all(
                isinstance(rs, RasterioSource) for rs in self.raster_sources)
            if not all_rasterio_sources:
                raise NotImplementedError(
                    'Non-identical extents are only '
                    'supported for RasterioSource raster sources.')

    @property
    def primary_source(self) -> RasterSource:
        """Primary sub-``RasterSource``"""
        return self.raster_sources[self.primary_source_idx]

    @property
    def dtype(self) -> np.dtype:
        return self.primary_source.dtype

    @property
    def crs_transformer(self) -> CRSTransformer:
        return self.primary_source.crs_transformer

    def _get_sub_chips(self, window: Box,
                       raw: bool = False) -> List[np.ndarray]:
        """If all extents are identical, simply retrieves chips from each sub
        raster source. Otherwise, follows the following algorithm
            - using pixel-coords window, get chip from the primary sub raster
            source
            - convert window to world coords using the CRS of the primary sub
            raster source
            - for each remaining sub raster source
                - convert world-coords window to pixel coords using the sub
                raster source's CRS
                - get chip from the sub raster source using this window;
                specify `out_shape` when reading to ensure shape matches
                reference chip from the primary sub raster source

        Args:
            window (Box): window to read, in pixel coordinates.
            raw (bool, optional): If True, uses RasterSource._get_chip.
                Otherwise, RasterSource.get_chip. Defaults to False.

        Returns:
            List[np.ndarray]: List of chips from each sub raster source.
        """

        def get_chip(
                rs: RasterSource,
                window: Box,
                out_shape: Optional[Tuple[int, int]] = None) -> np.ndarray:
            if raw:
                return rs._get_chip(window, out_shape=out_shape)
            return rs.get_chip(window, out_shape=out_shape)

        if self.all_extents_equal:
            sub_chips = [get_chip(rs, window) for rs in self.raster_sources]
        else:
            primary_rs = self.primary_source
            other_rses = [rs for rs in self.raster_sources if rs != primary_rs]

            primary_sub_chip = get_chip(primary_rs, window)
            out_shape = primary_sub_chip.shape[:2]
            world_window = primary_rs.crs_transformer.pixel_to_map(window)
            pixel_windows = [
                rs.crs_transformer.map_to_pixel(world_window)
                for rs in other_rses
            ]
            sub_chips = [
                get_chip(rs, w, out_shape=out_shape)
                for rs, w in zip(other_rses, pixel_windows)
            ]
            sub_chips.insert(self.primary_source_idx, primary_sub_chip)

        if self.force_same_dtype:
            dtype = sub_chips[self.primary_source_idx].dtype
            sub_chips = [chip.astype(dtype) for chip in sub_chips]

        return sub_chips

    def _get_chip(self, window: Box) -> np.ndarray:
        """Return the raw chip located in the window.

        Get raw chips from sub raster sources and concatenate them.

        Args:
            window: Box

        Returns:
            [height, width, channels] numpy array
        """
        sub_chips = self._get_sub_chips(window, raw=True)
        chip = np.concatenate(sub_chips, axis=-1)
        return chip

    def get_chip(self, window: Box) -> np.ndarray:
        """Return the transformed chip in the window.

        Get processed chips from sub raster sources (with their respective
        channel orders and transformations applied), concatenate them along the
        channel dimension, apply channel_order, followed by transformations.

        Args:
            window: Box

        Returns:
            np.ndarray with shape [height, width, channels]
        """
        sub_chips = self._get_sub_chips(window, raw=False)
        chip = np.concatenate(sub_chips, axis=-1)
        chip = chip[..., self.channel_order]

        for transformer in self.raster_transformers:
            chip = transformer.transform(chip, self.channel_order)

        return chip



from typing import TYPE_CHECKING
import logging

import numpy as np
from sklearn.metrics import confusion_matrix
from tqdm.auto import tqdm

from rastervision.core.evaluation import ClassEvaluationItem
from rastervision.core.evaluation import ClassificationEvaluation

if TYPE_CHECKING:
    from rastervision.core.data import (ClassConfig,
                                        SemanticSegmentationLabels)

log = logging.getLogger(__name__)


class SemanticSegmentationEvaluation(ClassificationEvaluation):
    """Evaluation for semantic segmentation."""

    def __init__(self, class_config: 'ClassConfig'):
        super().__init__()
        self.class_config = class_config

    def compute(self, gt_labels: 'SemanticSegmentationLabels',
                pred_labels: 'SemanticSegmentationLabels') -> None:
        self.reset()

        # compute confusion matrix
        null_class_id = self.class_config.null_class_id
        num_classes = len(self.class_config)
        labels = np.arange(num_classes)
        self.conf_mat = np.zeros((num_classes, num_classes))
        windows = pred_labels.get_windows()
        with tqdm(windows, delay=5, desc='Computing metrics') as bar:
            for window in bar:
                gt_arr = gt_labels.get_label_arr(window, null_class_id)
                pred_arr = pred_labels.get_label_arr(window, null_class_id)
                self.conf_mat += confusion_matrix(
                    gt_arr.ravel(), pred_arr.ravel(), labels=labels)

        for class_id, class_name in enumerate(self.class_config.names):
            eval_item = ClassEvaluationItem.from_multiclass_conf_mat(
                conf_mat=self.conf_mat,
                class_id=class_id,
                class_name=class_name)
            self.class_to_eval_item[class_id] = eval_item

        self.compute_avg()



from typing import TYPE_CHECKING, List
from abc import ABC, abstractmethod
import logging

from shapely.ops import unary_union
import geopandas as gpd

from rastervision.core.box import Box
from rastervision.core.data.utils import (
    remove_empty_features, split_multi_geometries, map_to_pixel_coords,
    pixel_to_map_coords, simplify_polygons, all_geoms_valid, geojson_to_geoms)

if TYPE_CHECKING:
    from shapely.geometry.base import BaseGeometry
    from rastervision.core.data import CRSTransformer, VectorTransformer

log = logging.getLogger(__name__)


class VectorSource(ABC):
    """A source of vector data."""

    def __init__(self,
                 crs_transformer: 'CRSTransformer',
                 vector_transformers: List['VectorTransformer'] = []):
        """Constructor.

        Args:
            crs_transformer: A ``CRSTransformer`` to convert
                between map and pixel coords. Normally this is obtained from a
                :class:`.RasterSource`.
            vector_transformers: ``VectorTransformers`` for transforming
                geometries. Defaults to ``[]``.
        """
        self.crs_transformer = crs_transformer
        self.vector_transformers = vector_transformers
        self._geojson = None
        self._extent = None

    def get_geojson(self, to_map_coords: bool = False) -> dict:
        """Return transformed GeoJSON.

        This makes the following transformations to the raw geojson:

        - converts to pixels coords (by default)
        - removes empty features
        - splits apart multi-geoms and geom collections into single geometries
        - buffers lines and points into Polygons

        Additionally, the transformations specified by all the
        VectorTransformers in vector_transformers are also applied.

        Args:
            to_map_coords: If true, will return GeoJSON in map coordinates.

        Returns:
            dict in GeoJSON format
        """
        if self._geojson is not None:
            return self._geojson

        geojson = self._get_geojson()
        geojson = sanitize_geojson(
            geojson, self.crs_transformer, to_map_coords=to_map_coords)

        for tf in self.vector_transformers:
            geojson = tf(geojson, crs_transformer=self.crs_transformer)

        self._geojson = geojson
        return geojson

    def get_geoms(self, to_map_coords: bool = False) -> List['BaseGeometry']:
        """Returns all geometries in the transformed GeoJSON as Shapely geoms.

        Args:
            to_map_coords: If true, will return geoms in map coordinates.

        Returns:
            List['BaseGeometry']: List of Shapely geoms.
        """
        geojson = self.get_geojson(to_map_coords=to_map_coords)
        return list(geojson_to_geoms(geojson))

    @abstractmethod
    def _get_geojson(self) -> dict:
        """Return raw GeoJSON."""
        pass

    def get_dataframe(self, to_map_coords: bool = False) -> gpd.GeoDataFrame:
        """Return geometries as a :class:`~geopandas.GeoDataFrame`."""
        geojson = self.get_geojson(to_map_coords=to_map_coords)
        df = gpd.GeoDataFrame.from_features(geojson)
        if len(df) == 0 and 'geometry' not in df.columns:
            df.loc[:, 'geometry'] = []
        return df

    @property
    def extent(self) -> Box:
        """Envelope of union of all geoms."""
        if self._extent is None:
            envelope = unary_union(self.get_geoms()).envelope
            self._extent = Box.from_shapely(envelope).to_int()
        return self._extent


def sanitize_geojson(geojson: dict,
                     crs_transformer: 'CRSTransformer',
                     to_map_coords: bool = False) -> dict:
    """Apply some basic transformations (listed below) to a GeoJSON.

    The following transformations are applied:

    1.  Removal of features without geometries.
    2.  Coordinate transformation to pixel coordinates.
    3.  Splitting of multi-part geometries e.g. MultiPolygon --> Polygons.
    4.  (Optional) If to_map_coords=true, transformation back to map
        coordinates.

    Args:
        geojson (dict): A GeoJSON-like mapping of a FeatureCollection.
        crs_transformer (CRSTransformer): A CRS transformer for coordinate
            transformation.
        to_map_coords (bool, optional): If True, transform geometries back to
            map coordinates before returning. Defaults to False.

    Returns:
        dict: Transformed FeatureCollection.
    """
    geojson = remove_empty_features(geojson)
    geojson = map_to_pixel_coords(geojson, crs_transformer)
    geojson = split_multi_geometries(geojson)
    geojson = simplify_polygons(geojson)
    if to_map_coords:
        geojson = pixel_to_map_coords(geojson, crs_transformer)
    if not all_geoms_valid(geojson):
        log.warn(f'Invalid geometries found in features in the GeoJSON.')
    return geojson



from typing import Dict

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.raster_transformer import (RasterTransformerConfig,
                                                       ReclassTransformer)


@register_config('reclass_transformer')
class ReclassTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.ReclassTransformer`."""

    mapping: Dict[int, int] = Field(
        ..., description=('The reclassification mapping.'))

    def build(self) -> ReclassTransformer:
        return ReclassTransformer(mapping=self.mapping)



from typing import TYPE_CHECKING, Iterable, Optional, Tuple
from rastervision.pipeline.config import register_config
from rastervision.core.evaluation.classification_evaluator_config import (
    ClassificationEvaluatorConfig)
from rastervision.core.evaluation.chip_classification_evaluator import (
    ChipClassificationEvaluator)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


@register_config('chip_classification_evaluator')
class ChipClassificationEvaluatorConfig(ClassificationEvaluatorConfig):
    """Configure a :class:`.ChipClassificationEvaluator`."""

    def build(self,
              class_config: 'ClassConfig',
              scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> ChipClassificationEvaluator:
        if scene_group is None:
            output_uri = self.get_output_uri()
        else:
            group_name, _ = scene_group
            output_uri = self.get_output_uri(group_name)

        return ChipClassificationEvaluator(class_config, output_uri)



"""Defines the abstract Labels class."""

from typing import TYPE_CHECKING, Any, Iterable
from abc import (ABC, abstractclassmethod, abstractmethod)

if TYPE_CHECKING:
    from rastervision.core.box import Box


class Labels(ABC):
    """A source-agnostic, in-memory representation of labels in a scene.

    This class can represent labels obtained via a ``LabelSource``, a
    ``LabelStore``, or directly from model predictions.
    """

    @abstractmethod
    def __add__(self, other):
        """Add labels to these labels.

        Returns a concatenation of this and the other labels.
        """
        pass

    @abstractmethod
    def filter_by_aoi(self, aoi_polygons):
        """Returns a copy of these labels filtered by a given set of AOI polygons

        Args:
          aoi_polygons - A list of AOI polygons to filter by, in pixel coordinates.
        """
        pass

    @abstractmethod
    def __eq__(self, other):
        pass

    @abstractmethod
    def __setitem__(self, key, value):
        pass

    @abstractclassmethod
    def make_empty(cls) -> 'Labels':
        """Instantiate an empty instance of this class.

        Returns:
            Labels: An object of the Label subclass on which this method is
                called.
        """
        pass

    @classmethod
    def from_predictions(cls, windows: Iterable['Box'],
                         predictions: Iterable[Any]) -> 'Labels':
        """Instantiate from windows and their corresponding predictions.

        This makes no assumptions about the type or format of the predictions.
        Subclasses should implement the __setitem__ method to correctly handle
        the predictions.

        Args:
            windows (Iterable[Box]): Boxes in pixel coords, specifying chips
                in the raster.
            predictions (Iterable[Any]): The model predictions for each chip
                specified by the windows.

        Returns:
            Labels: An object of the Label subclass on which this method is
                called.
        """
        labels = cls.make_empty()
        # If predictions is tqdm-wrapped, it needs to be the first arg to zip()
        # or the progress bar won't terminate with the correct count.
        for prediction, window in zip(predictions, windows):
            labels[window] = prediction
        return labels

    @abstractmethod
    def save(self, uri: str) -> None:
        """Save to file."""
        pass



from os.path import join
from typing import List, TYPE_CHECKING, Optional

from rastervision.pipeline.pipeline_config import PipelineConfig
from rastervision.core.data import (DatasetConfig, StatsTransformerConfig,
                                    LabelStoreConfig, SceneConfig)
from rastervision.core.utils.misc import Proportion
from rastervision.core.analyzer import StatsAnalyzerConfig
from rastervision.core.backend import BackendConfig
from rastervision.core.evaluation import EvaluatorConfig
from rastervision.core.analyzer import AnalyzerConfig
from rastervision.pipeline.config import (register_config, Field, Config)

if TYPE_CHECKING:
    from rastervision.core.backend.backend import Backend  # noqa


@register_config('predict_options')
class PredictOptions(Config):
    # TODO: predict_chip_sz and predict_batch_sz should probably be moved here
    pass


@register_config('rv_pipeline')
class RVPipelineConfig(PipelineConfig):
    """Configure an :class:`.RVPipeline`."""

    dataset: DatasetConfig = Field(
        ...,
        description=
        'Dataset containing train, validation, and optional test scenes.')
    backend: BackendConfig = Field(
        ..., description='Backend to use for interfacing with ML library.')
    evaluators: List[EvaluatorConfig] = Field(
        [],
        description=(
            'Evaluators to run during analyzer command. If list is empty '
            'the default evaluator is added.'))
    analyzers: List[AnalyzerConfig] = Field(
        [],
        description=
        ('Analyzers to run during analyzer command. A StatsAnalyzer will be added '
         'automatically if any scenes have a RasterTransformer.'))

    train_chip_sz: int = Field(
        300, description='Size of training chips in pixels.')
    predict_chip_sz: int = Field(
        300, description='Size of predictions chips in pixels.')
    predict_batch_sz: int = Field(
        8, description='Batch size to use during prediction.')
    chip_nodata_threshold: Proportion = Field(
        1,
        description='Discard chips where the proportion of NODATA values is '
        'greater than or equal to this value. Might result in false positives '
        'if there are many legitimate black pixels in the chip. Use with '
        'caution.')

    analyze_uri: Optional[str] = Field(
        None,
        description=
        'URI for output of analyze. If None, will be auto-generated.')
    chip_uri: Optional[str] = Field(
        None,
        description='URI for output of chip. If None, will be auto-generated.')
    train_uri: Optional[str] = Field(
        None,
        description='URI for output of train. If None, will be auto-generated.'
    )
    predict_uri: Optional[str] = Field(
        None,
        description=
        'URI for output of predict. If None, will be auto-generated.')
    eval_uri: Optional[str] = Field(
        None,
        description='URI for output of eval. If None, will be auto-generated.')
    bundle_uri: Optional[str] = Field(
        None,
        description='URI for output of bundle. If None, will be auto-generated.'
    )
    source_bundle_uri: Optional[str] = Field(
        None,
        description='If provided, the model will be loaded from this bundle '
        'for the train stage. Useful for fine-tuning.')

    def update(self):
        super().update()

        if self.analyze_uri is None:
            self.analyze_uri = join(self.root_uri, 'analyze')
        if self.chip_uri is None:
            self.chip_uri = join(self.root_uri, 'chip')
        if self.train_uri is None:
            self.train_uri = join(self.root_uri, 'train')
        if self.predict_uri is None:
            self.predict_uri = join(self.root_uri, 'predict')
        if self.eval_uri is None:
            self.eval_uri = join(self.root_uri, 'eval')
        if self.bundle_uri is None:
            self.bundle_uri = join(self.root_uri, 'bundle')

        self.dataset.update(pipeline=self)
        self.backend.update(pipeline=self)
        if not self.evaluators:
            self.evaluators.append(self.get_default_evaluator())
        for evaluator in self.evaluators:
            evaluator.update(pipeline=self)

        self._insert_analyzers()
        for analyzer in self.analyzers:
            analyzer.update(pipeline=self)

    def get_model_bundle_uri(self):
        return join(self.bundle_uri, 'model-bundle.zip')

    def _insert_analyzers(self):
        # Inserts StatsAnalyzer if it's needed because a RasterSource has a
        # StatsTransformer, but there isn't a StatsAnalyzer in the list of Analyzers.
        has_stats_transformer = False
        for s in self.dataset.all_scenes:
            for t in s.raster_source.transformers:
                if isinstance(t, StatsTransformerConfig):
                    has_stats_transformer = True

        has_stats_analyzer = False
        for a in self.analyzers:
            if isinstance(a, StatsAnalyzerConfig):
                has_stats_analyzer = True
                break

        if has_stats_transformer and not has_stats_analyzer:
            self.analyzers.append(StatsAnalyzerConfig())

    def get_default_label_store(self, scene: SceneConfig) -> LabelStoreConfig:
        """Returns a default LabelStoreConfig to fill in any missing ones."""
        raise NotImplementedError()

    def get_default_evaluator(self) -> EvaluatorConfig:
        """Returns a default EvaluatorConfig to use if one isn't set."""
        raise NotImplementedError()



"""Defines abstract base evaluation class for all tasks."""

from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from abc import (ABC, abstractmethod)
import copy
import json

import numpy as np

from rastervision.pipeline.file_system import str_to_file

if TYPE_CHECKING:
    from rastervision.core.evaluation import ClassEvaluationItem


class ClassificationEvaluation(ABC):
    """Base class for evaluating predictions for pipelines that have classes.

    Evaluations can be keyed, for instance, if evaluations happen per class.

    Attributes:
        class_to_eval_item (Dict[int, ClassEvaluationItem]): Mapping from class
            IDs to ``ClassEvaluationItem``s.
        scene_to_eval (Dict[str, ClassificationEvaluation]): Mapping from scene
            IDs to ``ClassificationEvaluation``s.
        avg_item (Optional[Dict[str, Any]]): Averaged evaluation over all
            classes.
        conf_mat (Optional[np.ndarray]): Confusion matrix.
    """

    def __init__(self):
        self.class_to_eval_item: Dict[int, 'ClassEvaluationItem']
        self.scene_to_eval: Dict[str, 'ClassificationEvaluation']
        self.avg_item: Optional[Dict[str, Any]]
        self.conf_mat: Optional[np.ndarray]
        self.reset()

    def reset(self):
        """Reset the Evaluation."""
        self.class_to_eval_item = {}
        self.scene_to_eval = {}
        self.avg_item = None
        self.conf_mat = None

    def to_json(self) -> Union[dict, list]:
        """Serialize to a dict or list.

        Returns:
            Union[dict, list]: Class-wise and (if available) scene-wise
                evaluations.
        """
        out = [item.to_json() for item in self.class_to_eval_item.values()]
        if self.avg_item:
            out.append(self.avg_item)

        if len(self.scene_to_eval) > 0:
            # append per scene evals
            out = {'overall': out}
            per_scene_evals = {
                scene_id: eval.to_json()
                for scene_id, eval in self.scene_to_eval.items()
            }
            out['per_scene'] = per_scene_evals

        return out

    def save(self, output_uri: str) -> None:
        """Save this Evaluation to a file.

        Args:
            output_uri: string URI for the file to write.
        """
        json_str = json.dumps(
            ensure_json_serializable(self.to_json()), indent=4)
        str_to_file(json_str, output_uri)

    def merge(self,
              other: 'ClassificationEvaluation',
              scene_id: Optional[str] = None) -> None:
        """Merge Evaluation for another Scene into this one.

        This is useful for computing the average metrics of a set of scenes.
        The results of the averaging are stored in this Evaluation.

        Args:
            other (ClassificationEvaluation): Evaluation to merge into this one
            scene_id (Optional[str], optional): ID of scene. If specified,
                (a copy of) ``other`` will be saved and be availabel in
                ``to_json()``'s output. Defaults to None.
        """
        if self.conf_mat is None:
            self.conf_mat = other.conf_mat
        else:
            self.conf_mat += other.conf_mat

        if len(self.class_to_eval_item) == 0:
            self.class_to_eval_item = other.class_to_eval_item
        else:
            for class_id, other_eval_item in other.class_to_eval_item.items():
                if class_id in self.class_to_eval_item:
                    self.class_to_eval_item[class_id].merge(other_eval_item)
                else:
                    self.class_to_eval_item[class_id] = other_eval_item

        self.compute_avg()

        if scene_id is not None:
            self.scene_to_eval[scene_id] = copy.deepcopy(other)

    def compute_avg(self) -> None:
        """Compute average metrics over all classes."""
        if len(self.class_to_eval_item) == 0:
            return
        class_evals = [
            eval_item.to_json()
            for eval_item in self.class_to_eval_item.values()
        ]
        # compute weighted averages of metrics
        class_counts = np.array([e['gt_count'] for e in class_evals])
        class_weights = class_counts / class_counts.sum()
        class_metrics = [e['metrics'] for e in class_evals]
        metric_names = class_metrics[0].keys()
        avg_metrics = {}
        for k in metric_names:
            metric_vals = np.array([m[k] for m in class_metrics], dtype=float)
            avg_metrics[k] = np.nan_to_num(metric_vals * class_weights).sum()

        # sum the counts
        gt_count = sum(e['gt_count'] for e in class_evals)
        pred_count = sum(e['pred_count'] for e in class_evals)
        count_error = sum(e['count_error'] for e in class_evals)

        self.avg_item = {
            'class_name': 'average',
            'metrics': avg_metrics,
            'gt_count': gt_count,
            'pred_count': pred_count,
            'count_error': count_error
        }
        if self.conf_mat is not None:
            cm = self.conf_mat
            self.avg_item['conf_mat'] = cm.tolist()
            self.avg_item['conf_mat_frac'] = (cm / cm.sum()).tolist()

    @abstractmethod
    def compute(self, ground_truth_labels, prediction_labels):
        """Compute metrics for a single scene.

        Args:
            ground_truth_labels: Ground Truth labels to evaluate against.
            prediction_labels: The predicted labels to evaluate.
        """
        pass


def ensure_json_serializable(obj: Any) -> dict:
    """Convert numpy types to JSON serializable equivalents."""
    if obj is None or isinstance(obj, (str, int, bool)):
        return obj
    if isinstance(obj, dict):
        return {k: ensure_json_serializable(v) for k, v in obj.items()}
    if isinstance(obj, (list, tuple)):
        return [ensure_json_serializable(o) for o in obj]
    if isinstance(obj, np.ndarray):
        return ensure_json_serializable(obj.tolist())
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, float):
        if np.isnan(obj):
            return None
        return float(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    return obj



from typing import TYPE_CHECKING, Iterable, Optional, Tuple
from os.path import join

from rastervision.pipeline.config import register_config, ConfigError, Field
from rastervision.core.analyzer import AnalyzerConfig, StatsAnalyzer

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig


@register_config('stats_analyzer')
class StatsAnalyzerConfig(AnalyzerConfig):
    """Configure a :class:`.StatsAnalyzer`.

    A :class:`.StatsAnalyzer` computes imagery statistics of scenes which can
    be used to normalize chips read from them.
    """

    output_uri: Optional[str] = Field(
        None,
        description='URI of directory where stats will be saved. '
        'Stats for a scene-group will be save in a JSON file at '
        '<output_uri>/<scene-group-name>/stats.json. If None, and this Config '
        'is part of an RVPipeline, this field will be auto-generated.')
    sample_prob: Optional[float] = Field(
        0.1,
        description=(
            'The probability of using a random window for computing statistics. '
            'If None, will use a sliding window.'))

    def update(self, pipeline: Optional['RVPipelineConfig'] = None) -> None:
        if pipeline is not None and self.output_uri is None:
            self.output_uri = join(pipeline.analyze_uri, 'stats')

    def validate_config(self):
        if self.sample_prob > 1 or self.sample_prob <= 0:
            raise ConfigError('sample_prob must be <= 1 and > 0')

    def build(self, scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> StatsAnalyzer:
        if scene_group is None:
            output_uri = join(self.output_uri, f'stats.json')
        else:
            group_name, _ = scene_group
            output_uri = join(self.output_uri, group_name, f'stats.json')
        return StatsAnalyzer(output_uri, sample_prob=self.sample_prob)

    def get_bundle_filenames(self):
        return ['stats.json']
